import os
import pickle
import uuid

from contextlib import contextmanager

from dagster import (
    check,
    ExecutionTargetHandle,
    ExpectationResult,
    Failure,
    Materialization,
    ModeDefinition,
    PipelineDefinition,
    RunConfig,
    SolidDefinition,
    TypeCheck,
)
from dagster.core.execution.api import scoped_pipeline_context
from dagster.core.execution.context_creation_pipeline import ResourcesStack
from dagster.core.definitions.dependency import SolidHandle
from dagster.loggers import colored_console_logger

from .context import DagstermillExecutionContext
from .errors import DagstermillError
from .logger import construct_sqlite_logger
from .serialize import PICKLE_PROTOCOL, read_value, write_value


class Manager:
    def __init__(self):
        self.handle = None
        self.pipeline_def = None
        self.solid_def = None
        self.in_pipeline = False
        self.marshal_dir = None
        self.context = None
        self.resources_stack = None

    @contextmanager
    def _setup_resources(self, pipeline_def, environment_config, run_config, log_manager):
        '''This context manager is a drop-in replacement for
        dagster.core.execution.context_creation_pipeline.create_resources. It uses the Manager's
        instance of ResourceStack to create resources, but does not tear them down when the
        context manager returns -- teardown must be managed manually using Manager.teardown().
        '''

        # pylint: disable=protected-access
        self.resources_stack = ResourcesStack(
            pipeline_def, environment_config, run_config, log_manager
        )
        yield self.resources_stack.create()

    def reconstitute_pipeline_context(
        self,
        output_log_path=None,
        marshal_dir=None,
        environment_dict=None,
        handle=None,
        run_config=None,
        solid_handle=None,
    ):
        '''Reconstitutes a context for dagstermill-managed execution.

        You'll see this function called to reconstruct a pipeline context within the ``injected
        parameters`` cell of a dagstermill output notebook. Users should not call this function
        interactively except when debugging output notebooks.

        Use :func:`dagstermill.get_context` in the ``parameters`` cell of your notebook to define a
        context for interactive exploration and development. This call will be replaced by one to
        :func:`dagstermill.reconstitute_pipeline_context` when the notebook is executed by
        dagstermill.
        '''
        check.opt_str_param(output_log_path, 'output_log_path')
        check.opt_str_param(marshal_dir, 'marshal_dir')
        environment_dict = check.opt_dict_param(environment_dict, 'environment_dict', key_type=str)
        check.inst_param(run_config, 'run_config', RunConfig)
        check.inst_param(handle, 'handle', ExecutionTargetHandle)
        check.inst_param(solid_handle, 'solid_handle', SolidHandle)

        pipeline_def = check.inst_param(
            handle.build_pipeline_definition(),
            'pipeline_def (from handle {handle_dict})'.format(handle_dict=handle.data._asdict()),
            PipelineDefinition,
        )

        solid_def = pipeline_def.get_solid(solid_handle)

        mode_def = pipeline_def.get_mode_definition(run_config.mode)
        shim_mode_def = ModeDefinition(
            name=mode_def.name,
            logger_defs=dict(
                mode_def.loggers, dagstermill=construct_sqlite_logger(output_log_path)
            ),
            resource_defs=mode_def.resource_defs,
        )

        pipeline_def = PipelineDefinition(
            pipeline_def.solid_defs,
            name=pipeline_def.name,
            description=pipeline_def.description,
            dependencies=pipeline_def.dependencies,
            mode_defs=[shim_mode_def],
            preset_defs=pipeline_def.preset_defs,
        )

        if 'loggers' not in environment_dict:
            environment_dict['loggers'] = {'dagstermill': {}}

        if 'dagstermill' not in environment_dict['loggers']:
            environment_dict['loggers']['dagstermill'] = {}

        self.marshal_dir = marshal_dir
        self.in_pipeline = True
        self.solid_def = solid_def
        self.pipeline_def = pipeline_def

        with scoped_pipeline_context(
            self.pipeline_def,
            environment_dict,
            run_config,
            scoped_resources_builder_cm=self._setup_resources,
        ) as pipeline_context:
            self.context = DagstermillExecutionContext(pipeline_context)

        return self.context

    def get_context(self, solid_config=None, mode_def=None, environment_dict=None):
        '''Get a dagstermill execution context for interactive exploration and development.

        Args:
            solid_config (Optional[Any]): If specified, this value will be made available on the
                context as its ``solid_config`` property.
            mode_def (Optional[:class:`dagster.ModeDefinition`]): If specified, defines the mode to
                use to construct the context. Specify this if you would like a context constructed
                with specific ``resource_defs`` or ``logger_defs``. By default, an ephemeral mode
                with a console logger will be constructed.
            environment_dict(Optional[dict]): The environment config dict with which to construct
                the context.

        Returns:
            :class:`dagstermill.DagstermillExecutionContext`
        '''
        check.opt_inst_param(mode_def, 'mode_def', ModeDefinition)
        environment_dict = check.opt_dict_param(environment_dict, 'environment_dict', key_type=str)

        solid_def = SolidDefinition(
            name='this_solid',
            input_defs=[],
            compute_fn=lambda *args, **kwargs: None,
            output_defs=[],
            description='Ephemeral solid constructed by dagstermill.get_context()',
        )

        if not mode_def:
            mode_def = ModeDefinition(logger_defs={'dagstermill': colored_console_logger})
            environment_dict['loggers'] = {'dagstermill': {}}

        pipeline_def = PipelineDefinition(
            [solid_def], mode_defs=[mode_def], name='ephemeral_dagstermill_pipeline'
        )

        run_config = RunConfig(mode=mode_def.name)

        self.in_pipeline = False
        self.solid_def = solid_def
        self.pipeline_def = pipeline_def

        with scoped_pipeline_context(
            self.pipeline_def,
            environment_dict,
            run_config,
            scoped_resources_builder_cm=self._setup_resources,
        ) as pipeline_context:
            self.context = DagstermillExecutionContext(pipeline_context, solid_config)

        return self.context

    def yield_result(self, value, output_name='result'):
        '''Yield a result directly from notebook code.

        When called interactively or in development, returns its input.

        Args:
            value (Any): The value to yield.
            output_name (Optional[str]): The name of the result to yield (default: ``'result'``).
        '''
        if not self.in_pipeline:
            return value

        # deferred import for perf
        import scrapbook

        if not self.solid_def.has_output(output_name):
            raise DagstermillError(
                'Solid {solid_name} does not have output named {output_name}'.format(
                    solid_name=self.solid_def.name, output_name=output_name
                )
            )

        runtime_type = self.solid_def.output_def_named(output_name).runtime_type

        out_file = os.path.join(self.marshal_dir, 'output-{}'.format(output_name))
        scrapbook.glue(output_name, write_value(runtime_type, value, out_file))

    def yield_event(self, dagster_event):
        '''Yield a dagster event directly from notebook code.

        When called interactively or in development, returns its input.

        Args:
            dagster_event (Union[:class:`dagster.Materialization`, :class:`dagster.ExpectationResult`, :class:`dagster.TypeCheck`, :class:`dagster.Failure`]):
                An event to yield back to Dagster.
        '''
        check.inst_param(
            dagster_event, 'dagster_event', (Materialization, ExpectationResult, TypeCheck, Failure)
        )

        if not self.in_pipeline:
            return dagster_event

        # deferred import for perf
        import scrapbook

        event_id = 'event-{event_uuid}'.format(event_uuid=str(uuid.uuid4()))
        out_file_path = os.path.join(self.marshal_dir, event_id)
        with open(out_file_path, 'wb') as fd:
            fd.write(pickle.dumps(dagster_event, PICKLE_PROTOCOL))

        scrapbook.glue(event_id, out_file_path)

    def teardown_resources(self):
        if self.resources_stack is not None:
            self.resources_stack.teardown()

    def load_parameter(self, input_name, input_value):
        input_def = self.solid_def.input_def_named(input_name)
        return read_value(input_def.runtime_type, input_value)


MANAGER_FOR_NOTEBOOK_INSTANCE = Manager()



'''Logging scheme for dagstermill.

Dagstermill notebooks are executed in a subprocess (opened by jupyter_client.manager.KernelManager).
We need to ship logs from the subprocess back to the parent process (executing the solid compute
logic), but since we don't manage the subprocess directly we can't use a Queue/QueueHandler-based
scheme. (Socket-based schemes might work, but getting these to play nicely with Docker and the wide
range of possible user networking setups sounds like a headache.)

There are a variety of persistent file-based queue projects in the Python ecosystem, but they are
written for multithreaded cases, not multiprocessing cases; Celery is too heavy for our use
case; and obviously we want to avoid implementing our own locking logic (perhaps on top of
filelock).

Instead we use a sqlite database as a queue. The elements of this setup are exhibited cleanly
in dagstermill_tests/test_logger.py:

1. Dagstermill solid compute logic running in parent process opens a NamedTemporaryFile for use as
   a sqlite db.
2. Parent process initializes the db table by calling dagstermill.logger.init_db. We rely on a
   sqlite autoincrement primary key to preserve the order of log messages.
3. Parent process starts a watcher thread that spins up a JsonSqlite3LogWatcher, which polls the
   sqlite db table for new logs until its is_done flag (a threading.Event) is set. New log messages
   are rehydrated using logging.makeLogRecord and handed directly to the logging.Handlers configured
   on the pipeline DagsterLogManager.
4. Child process is started by papermill/nbconvert machinery to execute notebook.
5. Dagstermill notebook manager logic running in child process creates a notebook-specific
   DagsterLogManager with a single-handler (dagstermill.logger.JsonSqlite3Handler) logger. This
   handler writes the internal dict representation of logging.LogRecords generated by the dagster
   logging machinery to the sqlite table as json.
6. Child process returns. Parent proess sets is_done flag.
7. Parent process waits for watcher process to complete logging.

'''
import copy
import json
import logging
import sqlite3
import sys
import threading
import time

from dagster import check, seven
from dagster.core.log_manager import DagsterLogManager
from dagster.utils.log import construct_single_handler_logger


CREATE_LOG_TABLE_STATEMENT = '''create table if not exists logs (
    timestamp integer primary key asc,
    json_log text
)
'''

INSERT_LOG_RECORD_STATEMENT = '''insert into logs (json_log) values (
    '{json_log}'
)
'''

RETRIEVE_LOG_RECORDS_STATEMENT = '''select * from logs
    where timestamp >= {timestamp}
    order by timestamp asc
'''

if sys.version_info < (3,):  # pragma: nocover
    EVENT_TYPE = threading._Event  # pylint: disable=no-member, protected-access
else:
    EVENT_TYPE = threading.Event


def init_db(sqlite_db_path):
    conn = sqlite3.connect(sqlite_db_path)
    cursor = conn.cursor()
    cursor.execute(CREATE_LOG_TABLE_STATEMENT)
    conn.commit()


class JsonSqlite3Handler(logging.Handler):
    def __init__(self, sqlite_db_path):
        check.str_param(sqlite_db_path, 'sqlite_db_path')

        self.sqlite_db_path = sqlite_db_path
        self.conn = sqlite3.connect(self.sqlite_db_path)
        self.cursor = self.conn.cursor()

        init_db(sqlite_db_path)

        super(JsonSqlite3Handler, self).__init__()

    def emit(self, record):
        try:
            log_dict = copy.copy(record.__dict__)
            self.cursor.execute(
                INSERT_LOG_RECORD_STATEMENT.format(json_log=seven.json.dumps(log_dict))
            )
            self.conn.commit()

        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


class JsonSqlite3LogWatcher(object):
    def __init__(self, sqlite_db_path, log_manager, is_done):
        check.str_param(sqlite_db_path, 'sqlite_db_path')
        check.inst_param(log_manager, 'log_manager', DagsterLogManager)
        check.inst_param(is_done, 'is_done', EVENT_TYPE)

        self.sqlite_db_path = sqlite_db_path
        self.conn = sqlite3.connect(self.sqlite_db_path)
        self.cursor = self.conn.cursor()
        self.next_timestamp = 0
        self.log_manager = log_manager
        self.is_done = is_done

    def watch(self):
        last_pass = False
        while True:
            res = self.cursor.execute(
                RETRIEVE_LOG_RECORDS_STATEMENT.format(timestamp=self.next_timestamp)
            ).fetchall()
            if res:
                self.next_timestamp = res[-1][0] + 1
                json_records = [r[1] for r in res]
                for json_record in json_records:
                    record = logging.makeLogRecord(json.loads(json_record))

                    for logger in self.log_manager.loggers:
                        for handler in logger.handlers:
                            # Because we're rehydrating the LogMessage, rather than passing
                            # through Logger._log again (which would obscure the original metadata)
                            # we need to filter for log level here
                            if handler.level <= record.levelno:
                                handler.handle(record)

            time.sleep(1)
            if last_pass:
                break
            if self.is_done.is_set():
                last_pass = True


def construct_sqlite_logger(sqlite_db_path):
    return construct_single_handler_logger(
        'dagstermill', 'debug', JsonSqlite3Handler(sqlite_db_path)
    )



import argparse
import sys

from setuptools import find_packages, setup


def get_version(name):
    version = {}
    with open("dagstermill/version.py") as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    if name == 'dagstermill':
        return version['__version__']
    elif name == 'dagstermill-nightly':
        return version['__nightly__']
    else:
        raise Exception('Shouldn\'t be here: bad package name {name}'.format(name=name))


parser = argparse.ArgumentParser()
parser.add_argument('--nightly', action='store_true')


def _do_setup(name='dagstermill'):
    setup(
        name=name,
        version=get_version(name),
        author='Elementl',
        license='Apache-2.0',
        packages=find_packages(exclude=['dagstermill_tests']),
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        install_requires=[
            # standard python 2/3 compatability things
            'dagster-pandas',
            'enum-compat==0.0.2',
            'future>=0.16.0, <0.17.0a0',
            'ipykernel>=4.9.0',
            'nteract-scrapbook>=0.2.0',
            'papermill>=1.0.0',
            'scikit-learn==0.20.3',
        ],
        entry_points={"console_scripts": ['dagstermill = dagstermill.cli:main']},
    )


if __name__ == '__main__':
    parsed, unparsed = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + unparsed
    if parsed.nightly:
        _do_setup('dagstermill-nightly')
    else:
        _do_setup('dagstermill')



import copy
import os
import pickle
import tempfile
import threading
import uuid

import nbformat
import papermill
import six

from papermill.engines import papermill_engines
from papermill.parameterize import _find_first_tagged_cell_index
from papermill.iorw import load_notebook_node, write_ipynb

from dagster import (
    check,
    ExecutionTargetHandle,
    EventMetadataEntry,
    InputDefinition,
    Materialization,
    OutputDefinition,
    Output,
    SolidDefinition,
    seven,
)
from dagster.core.errors import user_code_error_boundary
from dagster.core.execution.context.system import SystemComputeExecutionContext
from dagster.core.execution.context.compute import ComputeExecutionContext
from dagster.core.types.field_utils import check_user_facing_opt_field_param
from dagster.utils import mkdir_p

from .engine import DagstermillNBConvertEngine
from .errors import DagstermillExecutionError, DagstermillError
from .logger import init_db, JsonSqlite3LogWatcher
from .serialize import read_value, write_value
from .translator import DagsterTranslator, RESERVED_INPUT_NAMES


# This is based on papermill.parameterize.parameterize_notebook
# Typically, papermill injects the injected-parameters cell *below* the parameters cell
# but we want to *replace* the parameters cell, which is what this function does.
def replace_parameters(context, nb, parameters):
    '''Assigned parameters into the appropiate place in the input notebook

    Args:
        nb (NotebookNode): Executable notebook object
        parameters (dict): Arbitrary keyword arguments to pass to the notebook parameters.
    '''
    check.dict_param(parameters, 'parameters')

    # Copy the nb object to avoid polluting the input
    nb = copy.deepcopy(nb)

    # papermill method chooses translator based on kernel_name and language, but we just call the
    # DagsterTranslator to generate parameter content based on the kernel_name
    param_content = DagsterTranslator.codify(parameters)

    newcell = nbformat.v4.new_code_cell(source=param_content)
    newcell.metadata['tags'] = ['injected-parameters']

    param_cell_index = _find_first_tagged_cell_index(nb, 'parameters')
    injected_cell_index = _find_first_tagged_cell_index(nb, 'injected-parameters')
    if injected_cell_index >= 0:
        # Replace the injected cell with a new version
        before = nb.cells[:injected_cell_index]
        after = nb.cells[injected_cell_index + 1 :]
        check.int_value_param(param_cell_index, -1, 'param_cell_index')
        # We should have blown away the parameters cell if there is an injected-parameters cell
    elif param_cell_index >= 0:
        # Replace the parameter cell with the injected-parameters cell
        before = nb.cells[:param_cell_index]
        after = nb.cells[param_cell_index + 1 :]
    else:
        # Inject to the top of the notebook, presumably first cell includes dagstermill import
        context.log.debug(
            (
                'Executing notebook with no tagged parameters cell: injecting boilerplate in first '
                'cell.'
            )
        )
        before = nb.cells[:1]
        after = nb.cells[1:]

    nb.cells = before + [newcell] + after
    nb.metadata.papermill['parameters'] = parameters

    return nb


def get_papermill_parameters(compute_context, inputs, output_log_path):
    check.inst_param(compute_context, 'compute_context', SystemComputeExecutionContext)
    check.param_invariant(
        isinstance(compute_context.environment_dict, dict),
        'compute_context',
        'SystemComputeExecutionContext must have valid environment_dict',
    )
    check.dict_param(inputs, 'inputs', key_type=six.string_types)

    run_id = compute_context.run_id

    marshal_dir = '/tmp/dagstermill/{run_id}/marshal'.format(run_id=run_id)
    mkdir_p(marshal_dir)

    handle = ExecutionTargetHandle.get_handle(compute_context.pipeline_def)

    if not handle:
        raise DagstermillError(
            'Can\'t execute a dagstermill solid from a pipeline that wasn\'t instantiated using '
            'an ExecutionTargetHandle'
        )

    dm_handle_kwargs = handle.data._asdict()

    dm_handle_kwargs['pipeline_name'] = compute_context.pipeline_def.name

    dm_context_dict = {
        'output_log_path': output_log_path,
        'marshal_dir': marshal_dir,
        'environment_dict': compute_context.environment_dict,
    }

    dm_run_config_kwargs = {'run_id': run_id, 'mode': compute_context.mode_def.name}

    dm_solid_handle_kwargs = compute_context.solid_handle._asdict()

    parameters = {}

    input_def_dict = compute_context.solid_def.input_dict
    for input_name, input_value in inputs.items():
        assert (
            input_name not in RESERVED_INPUT_NAMES
        ), 'Dagstermill solids cannot have inputs named {input_name}'.format(input_name=input_name)
        runtime_type = input_def_dict[input_name].runtime_type
        parameter_value = write_value(
            runtime_type, input_value, os.path.join(marshal_dir, 'input-{}'.format(input_name))
        )
        parameters[input_name] = parameter_value

    parameters['__dm_context'] = seven.json.dumps(dm_context_dict)
    parameters['__dm_handle_kwargs'] = seven.json.dumps(dm_handle_kwargs)
    parameters['__dm_run_config_kwargs'] = seven.json.dumps(dm_run_config_kwargs)
    parameters['__dm_solid_handle_kwargs'] = seven.json.dumps(dm_solid_handle_kwargs)

    return parameters


def _dm_solid_compute(name, notebook_path):
    check.str_param(name, 'name')
    check.str_param(notebook_path, 'notebook_path')

    def _t_fn(compute_context, inputs):
        check.inst_param(compute_context, 'compute_context', ComputeExecutionContext)
        check.param_invariant(
            isinstance(compute_context.environment_dict, dict),
            'context',
            'SystemComputeExecutionContext must have valid environment_dict',
        )

        system_compute_context = compute_context.get_system_context()

        base_dir = '/tmp/dagstermill/{run_id}/'.format(run_id=compute_context.run_id)
        output_notebook_dir = os.path.join(base_dir, 'output_notebooks/')
        mkdir_p(output_notebook_dir)

        temp_path = os.path.join(
            output_notebook_dir, '{prefix}-out.ipynb'.format(prefix=str(uuid.uuid4()))
        )

        with tempfile.NamedTemporaryFile() as output_log_file:
            output_log_path = output_log_file.name
            init_db(output_log_path)

            # Scaffold the registration here
            nb = load_notebook_node(notebook_path)
            nb_no_parameters = replace_parameters(
                system_compute_context,
                nb,
                get_papermill_parameters(system_compute_context, inputs, output_log_path),
            )
            intermediate_path = os.path.join(
                output_notebook_dir, '{prefix}-inter.ipynb'.format(prefix=str(uuid.uuid4()))
            )
            write_ipynb(nb_no_parameters, intermediate_path)

            # Although the type of is_done is threading._Event in py2, not threading.Event,
            # it is still constructed using the threading.Event() factory
            is_done = threading.Event()

            def log_watcher_thread_target():
                log_watcher = JsonSqlite3LogWatcher(
                    output_log_path, system_compute_context.log, is_done
                )
                log_watcher.watch()

            log_watcher_thread = threading.Thread(target=log_watcher_thread_target)

            log_watcher_thread.start()

            with user_code_error_boundary(
                DagstermillExecutionError,
                lambda: (
                    'Error occurred during the execution of Dagstermill solid '
                    '{solid_name}: {notebook_path}'.format(
                        solid_name=name, notebook_path=notebook_path
                    )
                ),
            ):
                try:
                    papermill_engines.register('dagstermill', DagstermillNBConvertEngine)
                    papermill.execute_notebook(
                        intermediate_path, temp_path, engine_name='dagstermill', log_output=True
                    )
                except Exception as exc:
                    yield Materialization(
                        label='output_notebook',
                        description='Location of output notebook on the filesystem',
                        metadata_entries=[EventMetadataEntry.fspath(temp_path)],
                    )
                    raise exc
                finally:
                    is_done.set()
                    log_watcher_thread.join()

            # deferred import for perf
            import scrapbook

            output_nb = scrapbook.read_notebook(temp_path)

            system_compute_context.log.debug(
                'Notebook execution complete for {name}. Data is {data}'.format(
                    name=name, data=output_nb.scraps
                )
            )

            yield Materialization(
                label='output_notebook',
                description='Location of output notebook on the filesystem',
                metadata_entries=[EventMetadataEntry.fspath(temp_path)],
            )

            for (output_name, output_def) in system_compute_context.solid_def.output_dict.items():
                data_dict = output_nb.scraps.data_dict
                if output_name in data_dict:
                    value = read_value(output_def.runtime_type, data_dict[output_name])

                    yield Output(value, output_name)

            for key, value in output_nb.scraps.items():
                if key.startswith('event-'):
                    with open(value.data, 'rb') as fd:
                        yield pickle.loads(fd.read())

    return _t_fn


def define_dagstermill_solid(
    name,
    notebook_path,
    input_defs=None,
    output_defs=None,
    config_field=None,
    required_resource_keys=None,
):
    '''Wrap a Jupyter notebook in a solid.

    Arguments:
        name (str): The name of the solid.
        notebook_path (str): Path to the backing notebook.
        input_defs (Optional[list[:class:`dagster.InputDefinition`]]): The solid's inputs.
        output_defs (Optional[list[:class:`dagster.OutputDefinition`]]): The solid's outputs.
        required_resource_keys (Optional[set[str]]): The string names of any required resources.

    Returns:
        :class:`dagster.SolidDefinition`
    '''
    check.str_param(name, 'name')
    check.str_param(notebook_path, 'notebook_path')
    input_defs = check.opt_list_param(input_defs, 'input_defs', of_type=InputDefinition)
    output_defs = check.opt_list_param(output_defs, 'output_defs', of_type=OutputDefinition)
    required_resource_keys = check.opt_set_param(
        required_resource_keys, 'required_resource_keys', of_type=str
    )
    config_field = check_user_facing_opt_field_param(
        config_field, 'config_field', 'of a dagstermill solid named "{name}"'.format(name=name)
    )

    return SolidDefinition(
        name=name,
        input_defs=input_defs,
        compute_fn=_dm_solid_compute(name, notebook_path),
        output_defs=output_defs,
        config_field=config_field,
        required_resource_keys=required_resource_keys,
        description='This solid is backed by the notebook at {path}'.format(path=notebook_path),
        metadata={'notebook_path': notebook_path, 'kind': 'ipynb'},
    )



from dagster import check, RuntimeType, seven


PICKLE_PROTOCOL = 2


def is_json_serializable(value):
    try:
        seven.json.dumps(value)
        return True
    except TypeError:
        return False


def read_value(runtime_type, value):
    check.inst_param(runtime_type, 'runtime_type', RuntimeType)
    if runtime_type.is_scalar:
        return value
    elif runtime_type.is_any and is_json_serializable(value):
        return value
    else:
        return runtime_type.serialization_strategy.deserialize_from_file(value)


def write_value(runtime_type, value, target_file):
    check.inst_param(runtime_type, 'runtime_type', RuntimeType)
    if runtime_type.is_scalar:
        return value
    elif runtime_type.is_any and is_json_serializable(value):
        return value
    else:
        runtime_type.serialization_strategy.serialize_to_file(value, target_file)
        return target_file



import papermill

RESERVED_INPUT_NAMES = [
    '__dm_context',
    '__dm_dagster',
    '__dm_dagster_cli_load_handle',
    '__dm_dagster_core_definitions_dependency',
    '__dm_dagstermill',
    '__dm_handle',
    '__dm_handle_kwargs',
    '__dm_json',
    '__dm_raise_from',
    '__dm_run_config',
    '__dm_run_config_kwargs',
    '__dm_solid_handle_kwargs',
    '__dm_sys',
]

INJECTED_BOILERPLATE = '''
# Injected parameters

import json as __dm_json
import sys as __dm_sys

import dagster as __dm_dagster
import dagster.cli.load_handle as __dm_dagster_cli_load_handle
import dagster.core.definitions.dependency as __dm_dagster_core_definitions_dependency
import dagstermill as __dm_dagstermill

# six.raise_from
if __dm_sys.version_info[:2] == (3, 2):
    exec("""def __dm_raise_from(value, from_value):
    try:
        if from_value is None:
            raise value
        raise value from from_value
    finally:
        value = None
""")
elif __dm_sys.version_info[:2] > (3, 2):
    exec("""def __dm_raise_from(value, from_value):
    try:
        raise value from from_value
    finally:
        value = None
""")
else:
    def __dm_raise_from(value, from_value):
        raise value

__dm_run_config = __dm_dagster.RunConfig(**__dm_json.loads('{dm_run_config_kwargs}'))

try:
    __dm_handle = __dm_dagster_cli_load_handle.handle_for_pipeline_cli_args(
        __dm_json.loads('{dm_handle_kwargs}'),
        use_default_repository_yaml=False,        
    )
except (__dm_dagster.check.CheckError, __dm_dagster_cli_load_handle.CliUsageError) as err:
    __dm_raise_from(
        __dm_dagstermill.DagstermillError(
            'Cannot invoke a dagstermill solid from an in-memory pipeline that was not loaded '
            'from an ExecutionTargetHandle. Run this pipeline using dagit, the dagster CLI, '
            'through dagster-graphql, or in-memory after loading it through an '
            'ExecutionTargetHandle.'
        ),
        err,
    )

__dm_solid_handle = __dm_dagster_core_definitions_dependency.SolidHandle.from_dict(
    __dm_json.loads('{dm_solid_handle_kwargs}')
)

context = __dm_dagstermill._reconstitute_pipeline_context(
    run_config=__dm_run_config,
    handle=__dm_handle,
    solid_handle=__dm_solid_handle,
    **__dm_json.loads('{dm_context}')
)
'''


class DagsterTranslator(papermill.translators.PythonTranslator):
    @classmethod
    def codify(cls, parameters):
        assert '__dm_context' in parameters
        assert '__dm_handle_kwargs' in parameters
        assert '__dm_run_config_kwargs' in parameters
        assert '__dm_solid_handle_kwargs' in parameters

        content = INJECTED_BOILERPLATE.format(
            dm_context=parameters['__dm_context'],
            dm_handle_kwargs=parameters['__dm_handle_kwargs'],
            dm_run_config_kwargs=parameters['__dm_run_config_kwargs'],
            dm_solid_handle_kwargs=parameters['__dm_solid_handle_kwargs'],
        )

        for name, val in parameters.items():
            if name in RESERVED_INPUT_NAMES:
                continue
            dm_unmarshal_call = '__dm_dagstermill._load_parameter("{name}", {val})'.format(
                name=name, val='"{val}"'.format(val=val) if isinstance(val, str) else val
            )
            content += '{}\n'.format(cls.assign(name, dm_unmarshal_call))

        return content


papermill.translators.papermill_translators.register('python', DagsterTranslator)



import copy
import os

import click
from papermill.iorw import load_notebook_node, write_ipynb
import nbformat

from dagster.utils import safe_isfile, mkdir_p


def get_notebook_scaffolding():
    first_cell_source = '"import dagstermill"'

    starting_notebook_init = '''
    {{
    "cells": [
    {{
    "cell_type": "code",
    "execution_count": null,
    "metadata": {{}},
    "outputs": [],
    "source": [
        {first_cell_source}
    ]
    }},
    {{
    "cell_type": "code",
    "execution_count": null,
    "metadata": {{
        "tags": [
        "parameters"
        ]
    }},
    "outputs": [],
    "source": [
        "context = dagstermill.get_context()"
    ]
    }}
    ],
    "metadata": {{
    "celltoolbar": "Tags"
    }},
    "nbformat": 4,
    "nbformat_minor": 2
    }}'''
    return starting_notebook_init.format(first_cell_source=first_cell_source)


@click.command(name='register-notebook', help=('Registers repository in existing notebook'))
@click.option('--notebook', '-note', type=click.STRING, help='Path to notebook')
def retroactively_scaffold_notebook(notebook):
    execute_retroactive_scaffold(notebook)


def execute_retroactive_scaffold(notebook_path):
    nb = load_notebook_node(notebook_path)
    new_nb = copy.deepcopy(nb)

    import_cell_source = 'import dagstermill'
    import_cell = nbformat.v4.new_code_cell(source=import_cell_source)

    parameters_cell_source = 'context = dagstermill.get_context()'
    parameters_cell = nbformat.v4.new_code_cell(source=parameters_cell_source)
    parameters_cell.metadata['tags'] = ['parameters']

    new_nb.cells = [import_cell, parameters_cell] + nb.cells
    write_ipynb(new_nb, notebook_path)


@click.command(name='create-notebook', help=('Creates new dagstermill notebook.'))
@click.option('--notebook', '-note', type=click.STRING, help="Name of notebook")
@click.option(
    '--force-overwrite',
    is_flag=True,
    help="Will force overwrite any existing notebook or file with the same name.",
)
def create_notebook(notebook, force_overwrite):
    execute_create_notebook(notebook, force_overwrite)


def execute_create_notebook(notebook, force_overwrite):
    notebook_path = os.path.join(
        os.getcwd(), notebook if notebook.endswith('.ipynb') else notebook + ".ipynb"
    )

    notebook_dir = os.path.dirname(notebook_path)
    mkdir_p(notebook_dir)

    if not force_overwrite and safe_isfile(notebook_path):
        click.confirm(
            (
                'Warning, {notebook_path} already exists and continuing '
                'will overwrite the existing notebook. '
                'Are you sure you want to continue?'
            ).format(notebook_path=notebook_path),
            abort=True,
        )

    with open(notebook_path, 'w') as f:
        f.write(get_notebook_scaffolding())
        click.echo("Created new dagstermill notebook at {path}".format(path=notebook_path))


def create_dagstermill_cli():
    group = click.Group(name="dagstermill")
    group.add_command(create_notebook)
    group.add_command(retroactively_scaffold_notebook)
    return group


def main():
    cli = create_dagstermill_cli()
    # click magic
    cli(obj={})  # pylint:disable=E1120
